//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "UnitMain.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TFormMain *FormMain;



int tile_side;

unsigned short *map;
int map_width;
int map_height;

unsigned char *tileset;
int tile_count;
int tile_count_allocated;

int tile_cursor_a;
int tile_cursor_b;

int color;

int map_scale;
int tileset_scale;



void __fastcall TFormMain::MapToTileset(void)
{
	int i,j,k,l,ptr,found,mptr,size;
	unsigned char *src,*tile;

	map_width =ImageMap->Picture->Bitmap->Width/tile_side;
	map_height=ImageMap->Picture->Bitmap->Height/tile_side;

	if(map) free(map);

	size=map_width*map_height*sizeof(unsigned short);

	map=(unsigned short*)malloc(size);
	memset(map,0,size);

	if(tileset) free(tileset);

	tile_count=0;
	tile_count_allocated=1024;

	size=tile_count_allocated*tile_side*tile_side;

	tileset=(unsigned char*)malloc(size);
	memset(tileset,0,size);

	size=tile_side*tile_side;

	tile=(unsigned char*)malloc(size);
	memset(tile,0,size);

	mptr=0;

	for(i=0;i<map_height;++i)
	{
		for(j=0;j<map_width;++j)
		{
			ptr=0;

			for(k=0;k<tile_side;++k)
			{
				src=(unsigned char*)ImageMap->Picture->Bitmap->ScanLine[i*tile_side+k]+j*tile_side;

				for(l=0;l<tile_side;++l) tile[ptr++]=*src++;
			}

			found=-1;

			for(k=0;k<tile_count;++k)
			{
				if(!memcmp(&tileset[k*tile_side*tile_side],tile,tile_side*tile_side))
				{
					found=k;
					break;
				}
			}

			if(found>=0)
			{
				map[mptr++]=found;
			}
			else
			{
				map[mptr++]=tile_count;

				memcpy(&tileset[tile_count*tile_side*tile_side],tile,tile_side*tile_side);

				++tile_count;

				if(tile_count>=tile_count_allocated)
				{
					tile_count_allocated+=1024;

					tileset=(unsigned char*)realloc(tileset,tile_count_allocated*tile_side*tile_side);
				}
			}
		}
	}

	free(tile);
}



void __fastcall TFormMain::TilesetToMap(bool cursor,bool repaint)
{
	int i,j,k,l,ptr,mptr,tile,sel_a,sel_b,col1,col2;
	unsigned char *dst;
	bool solo,tile_cursor,tile_normal,tile_empty;

	solo=false;
	sel_a=-1;
	sel_b=-1;

	if(cursor)
	{
		if(!SpeedButtonTileANone->Down) sel_a=tile_cursor_a;
		if(!SpeedButtonTileBNone->Down) sel_b=tile_cursor_b;

		if(SpeedButtonTileASolo->Down||SpeedButtonTileBSolo->Down) solo=true;
	}

	mptr=0;

	for(i=0;i<map_height;++i)
	{
		for(j=0;j<map_width;++j)
		{
			tile=map[mptr++];

			ptr=tile*tile_side*tile_side;

			tile_normal=false;
			tile_empty =false;
			tile_cursor=false;

			if(solo)
			{
				if(tile==sel_a||tile==sel_b) tile_normal=true; else tile_empty=true;
			}
			else
			{
				if(tile==sel_a||tile==sel_b) tile_cursor=true; else tile_normal=true;
			}

			if(tile_normal)
			{
				for(k=0;k<tile_side;++k)
				{
					dst=(unsigned char*)ImageMap->Picture->Bitmap->ScanLine[i*tile_side+k]+j*tile_side;

					memcpy(dst,&tileset[ptr],tile_side);

					ptr+=tile_side;
				}
			}

			if(tile_cursor)
			{
				if(tile==sel_a&&tile==sel_b)
				{
					col1=255;
					col2=254;
				}
				else
				{
					col1=(tile==sel_a?255:254);
					col2=col1;
				}

				for(k=0;k<tile_side;++k)
				{
					dst=(unsigned char*)ImageMap->Picture->Bitmap->ScanLine[i*tile_side+k]+j*tile_side;

					for(l=0;l<tile_side;l+=2)
					{
						if(k&1)
						{
							*dst++=((l&2)?col2:col1);
							*dst++=tileset[ptr+1];
						}
						else
						{
							*dst++=tileset[ptr];
							*dst++=((l&2)?col2:col1);
						}

						ptr+=2;
					}
				}
			}

			if(tile_empty)
			{
				for(k=0;k<tile_side;++k)
				{
					dst=(unsigned char*)ImageMap->Picture->Bitmap->ScanLine[i*tile_side+k]+j*tile_side;

					for(l=0;l<tile_side;++l)
					{
						if((k^l)&4)
						{
							*dst++=252;
						}
						else
						{
							*dst++=253;
						}
					}
				}
			}
		}
	}

	if(repaint) ImageMap->Repaint();
}



void __fastcall TFormMain::RenderTileset(void)
{
	int k,l,x,y,ptr,tile,col1,col2;
	unsigned char *dst;
	int tiles_wdt;

	ScrollBoxTileset->Width =PanelTools->Width -8-ScrollBoxTileset->Left;
	ScrollBoxTileset->Height=PanelTools->Height-8-ScrollBoxTileset->Top;

	tiles_wdt=(ScrollBoxTileset->Width-GetSystemMetrics(SM_CYVSCROLL))/(tile_side*tileset_scale);

	if(tiles_wdt<1) tiles_wdt=1;

	ImageTileset->Width =tiles_wdt*tile_side*tileset_scale;
	ImageTileset->Height=(tile_count+tiles_wdt-1)/tiles_wdt*tile_side*tileset_scale;

	ImageTileset->Picture->Bitmap->SetSize(ImageTileset->Width/tileset_scale,ImageTileset->Height/tileset_scale);
	ImageTileset->Picture->Bitmap->PixelFormat=pf8bit;

	ImageTileset->Picture->Bitmap->Palette=ImageMap->Picture->Bitmap->Palette;

	ImageTileset->Picture->Bitmap->Canvas->FillRect(TRect(0,0,ImageTileset->Picture->Bitmap->Width,ImageTileset->Picture->Bitmap->Height));

	x=0;
	y=0;

	for(tile=0;tile<tile_count;++tile)
	{
		ptr=tile*tile_side*tile_side;

		if(tile!=tile_cursor_a&&tile!=tile_cursor_b)
		{
			for(k=0;k<tile_side;++k)
			{
				dst=(unsigned char*)ImageTileset->Picture->Bitmap->ScanLine[y+k]+x;

				memcpy(dst,&tileset[ptr],tile_side);

				ptr+=tile_side;
			}
		}
		else
		{
			if(tile==tile_cursor_a&&tile==tile_cursor_b)
			{
				col1=255;
				col2=254;
			}
			else
			{
				col1=(tile==tile_cursor_a?255:254);
				col2=col1;
			}

			for(k=0;k<tile_side;++k)
			{
				dst=(unsigned char*)ImageTileset->Picture->Bitmap->ScanLine[y+k]+x;

				for(l=0;l<tile_side;l+=2)
				{
					if(k&1)
					{
						*dst++=((l&2)?col2:col1);
						*dst++=tileset[ptr+1];
					}
					else
					{
						*dst++=tileset[ptr];
						*dst++=((l&2)?col2:col1);
					}

					ptr+=2;
				}
			}
		}

		x+=tile_side;

		if(x>=ImageTileset->Picture->Bitmap->Width)
		{
			x=0;
			y+=tile_side;

			if(y>=ImageTileset->Picture->Bitmap->Height) break;
		}
	}

	ImageTileset->Repaint();
}



void __fastcall TFormMain::RenderTile(TImage *image,int tile)
{
	int i,j,ptr;
	unsigned char *dst;

	image->Picture->Bitmap->SetSize(tile_side,tile_side);
	image->Picture->Bitmap->PixelFormat=pf8bit;
	image->Picture->Bitmap->Palette=ImageTileset->Picture->Bitmap->Palette;

	ptr=tile*tile_side*tile_side;

	for(i=0;i<tile_side;++i)
	{
		for(j=0;j<tile_side;++j)
		{
			dst=(unsigned char*)image->Picture->Bitmap->ScanLine[i]+j;

			*dst=tileset[ptr++];
		}
	}

	image->Repaint();
}



void __fastcall TFormMain::TilesetFreqSort(void)
{
	int i,j,temp,pp,tile;
	int *sort,*freq;
	unsigned char *chrt;

	sort=(int*)malloc(tile_count*sizeof(int));
	freq=(int*)malloc(tile_count*sizeof(int));
	chrt=(unsigned char*)malloc(tile_count*tile_side*tile_side);

	for(i=0;i<tile_count;++i)
	{
		sort[i]=i;
		freq[i]=0;
	}

	for(i=0;i<map_width*map_height;++i) ++freq[map[i]];

	for(i=0;i<tile_count-1;++i)
	{
		for(j=0;j<tile_count-i-1;++j)
		{
			if(freq[j]<freq[j+1])
			{
				temp=freq[j];
				freq[j]=freq[j+1];
				freq[j+1]=temp;

				temp=sort[j];
				sort[j]=sort[j+1];
				sort[j+1]=temp;
			}
		}
	}

	for(i=0;i<map_width*map_height;++i)
	{
		for(j=0;j<tile_count;++j)
		{
			tile=map[i];

			if(sort[j]==tile)
			{
				map[i]=j;
				break;
			}
		}
	}

	memcpy(chrt,tileset,tile_count*tile_side*tile_side);

	pp=0;

	for(i=0;i<tile_count;++i)
	{
		memcpy(&tileset[pp],&chrt[sort[i]*tile_side*tile_side],tile_side*tile_side);

		pp+=tile_side*tile_side;
	}

	free(sort);
	free(freq);
	free(chrt);
}



void __fastcall TFormMain::UpdateInfo(void)
{
	int i,len,cur1,cur2;
	int *freq;

	len=tile_count;

	cur1=tile_cursor_a;
	cur2=tile_cursor_b;

	if(!len)
	{
		len=1;

		cur1=0;
		cur2=0;
	}

	freq=(int*)malloc(len*sizeof(int));

	for(i=0;i<len;++i) freq[i]=0;

	for(i=0;i<map_width*map_height;++i) ++freq[map[i]];

	LabelInfo->Caption=IntToStr(tile_count)+" tiles";

	LabelInfoTileA->Caption="Id:"+IntToStr(cur1)+" ("+IntToStr(freq[cur1])+" entries)";
	LabelInfoTileB->Caption="Id:"+IntToStr(cur2)+" ("+IntToStr(freq[cur2])+" entries)";

	ImageMap->Width =ImageMap->Picture->Bitmap->Width *map_scale;
	ImageMap->Height=ImageMap->Picture->Bitmap->Height*map_scale;

	free(freq);
}



void __fastcall TFormMain::ChangeTile(int &cursor,int tile)
{
	if(tile>=tile_count) return;

	cursor=tile;

	TilesetToMap(true,true);
	RenderTileset();
	RenderTile(ImageTileA,tile_cursor_a);
	RenderTile(ImageTileB,tile_cursor_b);
	UpdateInfo();
}



void __fastcall TFormMain::UpdateAll(void)
{
	TilesetToMap(false,false);
	MapToTileset();
	TilesetFreqSort();
	RenderTileset();
	RenderTile(ImageTileA,tile_cursor_a);
	RenderTile(ImageTileB,tile_cursor_b);
	TilesetToMap(true,true);
	UpdateInfo();
}



void __fastcall TFormMain::UpdateCursorColors(TImage *image)
{
	PALETTEENTRY pal[256];
	RGBQUAD qpal[256];
	int i;

	GetPaletteEntries(image->Picture->Bitmap->Palette,0,256,pal);

	pal[255].peBlue =(ShapeA->Brush->Color>>16)&0xff;
	pal[255].peGreen=(ShapeA->Brush->Color>>8)&0xff;
	pal[255].peRed  =(ShapeA->Brush->Color)&0xff;

	pal[254].peBlue =(ShapeB->Brush->Color>>16)&0xff;
	pal[254].peGreen=(ShapeB->Brush->Color>>8)&0xff;
	pal[254].peRed  =(ShapeB->Brush->Color)&0xff;

	pal[253].peBlue =0xc0;
	pal[253].peGreen=0xc0;
	pal[253].peRed  =0xc0;

	pal[252].peBlue =0xf0;
	pal[252].peGreen=0xf0;
	pal[252].peRed  =0xf0;

	for(i=0;i<256;++i)
	{
		qpal[i].rgbBlue =pal[i].peBlue;
		qpal[i].rgbGreen=pal[i].peGreen;
		qpal[i].rgbRed  =pal[i].peRed;
	}

	SetPaletteEntries(image->Picture->Bitmap->Palette,0,256,pal);
	SetDIBColorTable(image->Picture->Bitmap->Canvas->Handle,0,256,qpal);
}



void __fastcall TFormMain::ImportBMP(AnsiString filename)
{
	SaveDialogExport->FileName=filename;

	ImageMap->Picture->LoadFromFile(filename);

	ImageMap->Picture->Bitmap->PixelFormat=pf8bit;

	UpdateCursorColors(ImageMap);

	map_width=0;
	map_height=0;
	tile_count=0;

	tile_cursor_a=0;
	tile_cursor_b=0;

	UpdateAll();
}



//---------------------------------------------------------------------------
__fastcall TFormMain::TFormMain(TComponent* Owner)
: TForm(Owner)
{
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::FormDestroy(TObject *Sender)
{
	if(map) free(map);

	if(tileset) free(tileset);
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::ImageTilesetMouseDown(TObject *Sender,
TMouseButton Button, TShiftState Shift, int X, int Y)
{
	ImageTilesetMouseMove(Sender,Shift,X,Y);
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::ImageMapMouseDown(TObject *Sender, TMouseButton Button,
TShiftState Shift, int X, int Y)
{
	ImageMapMouseMove(Sender,Shift,X,Y);
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::CheckBoxShowSelectedClick(TObject *Sender)
{
	TilesetToMap(true,true);
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::ComboBoxSizeChange(TObject *Sender)
{
	int tile_side_new;

	if(ComboBoxSize->ItemIndex==0) tile_side_new=8; else tile_side_new=16;

	TilesetToMap(false,false);

	tile_side=tile_side_new;
	tile_cursor_a=0;
	tile_cursor_b=0;

	MapToTileset();
	TilesetFreqSort();
	RenderTileset();
	RenderTile(ImageTileA,tile_cursor_a);
	RenderTile(ImageTileB,tile_cursor_b);
	TilesetToMap(true,true);
	UpdateInfo();
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::ImageMapMouseMove(TObject *Sender, TShiftState Shift,
int X, int Y)
{
	int tx,ty,tile;

	ty=Y/map_scale/tile_side;
	tx=X/map_scale/tile_side;

	if(tx<0||tx>=map_width||ty<0||ty>=map_height) return;

	tile=map[ty*map_width+tx];

	if(Shift.Contains(ssLeft )) ChangeTile(tile_cursor_a,tile);
	if(Shift.Contains(ssRight)) ChangeTile(tile_cursor_b,tile);
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::ImageTilesetMouseMove(TObject *Sender,
TShiftState Shift, int X, int Y)
{
	int tile;

	if(X<0||X>=ImageTileset->Width||Y<0||Y>=ImageTileset->Height) return;

	tile=Y/(tile_side*tileset_scale)*(ImageTileset->Width/(tile_side*tileset_scale))+X/(tile_side*tileset_scale);

	if(Shift.Contains(ssLeft )) ChangeTile(tile_cursor_a,tile);
	if(Shift.Contains(ssRight)) ChangeTile(tile_cursor_b,tile);
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::ImageTileAMouseDown(TObject *Sender,
TMouseButton Button, TShiftState Shift, int X, int Y)
{
	ImageTileAMouseMove(Sender,Shift,X,Y);
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::ImageTileAMouseMove(TObject *Sender,
TShiftState Shift, int X, int Y)
{
	int px,py,off,tile;

	if(Sender==ImageTileA) tile=tile_cursor_a; else if(Sender==ImageTileB) tile=tile_cursor_b; else return;

	px=X/(ImageTileA->Width /tile_side);
	py=Y/(ImageTileA->Height/tile_side);

	if(px<0||py<0||px>=tile_side||py>=tile_side) return;

	off=tile*tile_side*tile_side+py*tile_side+px;

	if(Shift.Contains(ssLeft))
	{
		tileset[off]=color;

		RenderTile(ImageTileA,tile_cursor_a);
		RenderTile(ImageTileB,tile_cursor_b);
	}

	if(Shift.Contains(ssRight))
	{
		color=tileset[off];
	}
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::ImageTileAMouseUp(TObject *Sender,
TMouseButton Button, TShiftState Shift, int X, int Y)
{
	UpdateAll();
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::FormCreate(TObject *Sender)
{
	ImageTileA->Picture->Bitmap->ReleasePalette();
	ImageTileB->Picture->Bitmap->ReleasePalette();

	DoubleBuffered=true;
	PanelTools->DoubleBuffered=true;

	tile_side=16;

	map=NULL;
	map_width=0;
	map_height=0;

	tileset=NULL;
	tile_count=0;
	tile_count_allocated=0;

	tile_cursor_a=0;
	tile_cursor_b=0;

	map_scale=2;
	tileset_scale=2;

	UpdateInfo();

	if(ParamStr(1)!="") ImportBMP(ParamStr(1));
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::SpeedButtonMap2xClick(TObject *Sender)
{
	int n;

	n=-1;

	if(SpeedButtonMap2x->Down) n=2;
	if(SpeedButtonMap3x->Down) n=3;
	if(SpeedButtonMap4x->Down) n=4;

	if(map_scale!=n)
	{
		map_scale=n;

		UpdateAll();
	}
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::SpeedButtonTileset2xClick(TObject *Sender)
{
	int n;

	n=-1;

	if(SpeedButtonTileset2x->Down) n=2;
	if(SpeedButtonTileset3x->Down) n=3;
	if(SpeedButtonTileset4x->Down) n=4;

	if(tileset_scale!=n)
	{
		tileset_scale=n;

		UpdateAll();
	}
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::SpeedButtonImportClick(TObject *Sender)
{
	if(!OpenDialogImport->Execute()) return;

	ImportBMP(OpenDialogImport->FileName);
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::SpeedButtonExportClick(TObject *Sender)
{
	if(!SaveDialogExport->Execute()) return;

	TilesetToMap(false,true);

	ImageMap->Picture->SaveToFile(SaveDialogExport->FileName);	
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::SpeedButtonReplaceABClick(TObject *Sender)
{
	int i,tile_src,tile_dst;
	unsigned char *temp;

	//replace one tile id with another in the map

	if(Sender==SpeedButtonReplaceAB)
	{
		tile_src=tile_cursor_b;
		tile_dst=tile_cursor_a;
	}
	else
	{
		tile_src=tile_cursor_a;
		tile_dst=tile_cursor_b;
	}

	for(i=0;i<map_width*map_height;++i) if(map[i]==tile_src) map[i]=tile_dst;

	//remember graphics of the replaced tile

	temp=(unsigned char*)malloc(tile_side*tile_side);

	memcpy(temp,&tileset[tile_dst*tile_side*tile_side],tile_side*tile_side);

	//reconvert the map and cleanup the tileset

	TilesetToMap(false,false);
	MapToTileset();
	TilesetFreqSort();

	//search for the replaced tile graphics new location, put cursor there

	for(i=0;i<tile_count;++i)
	{
		if(!memcmp(temp,&tileset[i*tile_side*tile_side],tile_side*tile_side))
		{
			tile_cursor_a=i;
			tile_cursor_b=i;

			break;
		}
	}

	free(temp);

	UpdateAll();
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::FormResize(TObject *Sender)
{
	UpdateAll();
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::SpeedButtonTileACursorClick(TObject *Sender)
{
	UpdateAll();
}
//---------------------------------------------------------------------------

void __fastcall TFormMain::ShapeAMouseDown(TObject *Sender, TMouseButton Button,
      TShiftState Shift, int X, int Y)
{
	ColorDialogCursor->Color=((TShape*)Sender)->Brush->Color;

	ColorDialogCursor->Execute();

	((TShape*)Sender)->Brush->Color=ColorDialogCursor->Color;

	UpdateCursorColors(ImageMap);
	UpdateCursorColors(ImageTileset);

	UpdateAll();
}
//---------------------------------------------------------------------------


